class ContextBuilder:
    def __init__(self):
        self.context_list = ["https://www.w3.org/ns/activitystreams"]
        self.additional_contexts = {}

    def add(self, *args, **kwargs):
        self.additional_contexts = {**self.additional_contexts, **kwargs}
        self.context_list += args
        return self

    def absorb(self, obj):
        if not isinstance(obj["@context"], str):
            for item in obj["@context"]:
                if isinstance(item, dict):
                    self.add(**item)
                elif item != "https://www.w3.org/ns/activitystreams":
                    self.context_list.append(item)

        del obj["@context"]

        return obj

    def build(self):
        if len(self.additional_contexts) == 0:
            if len(self.context_list) == 1:
                return self.context_list[0]
            return self.context_list

        return self.context_list + [self.additional_contexts]



from .accept_builder import AcceptBuilder
from .create_builder import CreateBuilder
from .delete_builder import DeleteBuilder
from .follow_builder import FollowBuilder
from .like_builder import LikeBuilder
from .undo_builder import UndoBuilder


def build_follow(domain: str, actor: str, tofollow: str) -> FollowBuilder:
    return FollowBuilder(domain, actor, tofollow)


def build_accept(account: str, obj: dict) -> AcceptBuilder:
    return AcceptBuilder(account, obj)


def build_create(*args) -> CreateBuilder:
    return CreateBuilder(*args)


def build_delete(actor, object_id):
    return DeleteBuilder(actor, object_id)


def build_like(actor, object_id):
    return LikeBuilder(actor, object_id)


def build_undo(obj: dict) -> UndoBuilder:
    return UndoBuilder(obj)



from bovine.types import Visibility

from .common import build_context


class ActorBuilder:
    def __init__(self, name: str, actor_type: str = "Person"):
        self.name = name
        self.display_name = name
        self.actor_type = actor_type
        self.account_url: str | None = None
        self.public_key: str | None = None
        self.public_key_name: str | None = None
        self.context_builder = build_context()
        self.event_source = None
        self.proxy_url = None
        self.followers = None
        self.following = None

    def with_account_url(self, account_url: str):
        self.account_url = account_url
        self.inbox = self.account_url + "/inbox"
        self.outbox = self.account_url + "/outbox"
        self.event_source = self.account_url + "/serverSentEvents"
        return self

    def with_inbox(self, inbox: str):
        self.inbox = inbox
        return self

    def with_outbox(self, outbox: str):
        self.outbox = outbox
        return self

    def with_follow(self, followers, following):
        self.followers = followers
        self.following = following
        return self

    def with_event_source(self, event_source):
        self.event_source = event_source
        return self

    def with_proxy_url(self, proxy_url):
        self.proxy_url = proxy_url
        return self

    def with_public_key(self, public_key: str, key_name: str = "main-key"):
        self.public_key = public_key
        self.public_key_name = key_name
        self.context_builder.add("https://w3id.org/security/v1")
        return self

    def build(self, visibility=Visibility.PUBLIC):
        return {
            "@context": self.context_builder.build(),
            "name": self.display_name,
            "preferredUsername": self.name,
            "type": self.actor_type,
            **self._build_account_urls(visibility),
            **self._build_private_key(),
        }

    def _build_context(self):
        if self.public_key:
            return [
                "https://www.w3.org/ns/activitystreams",
                "https://w3id.org/security/v1",
            ]

        return "https://www.w3.org/ns/activitystreams"

    def _build_account_urls(self, visibility):
        if not self.account_url:
            return {}

        if visibility == Visibility.WEB:
            return {"id": self.account_url}

        result = {
            "id": self.account_url,
            "inbox": self.inbox,
            "outbox": self.outbox,
        }

        if visibility == Visibility.OWNER:
            result["endpoints"] = {}
            if self.event_source:
                result["endpoints"]["eventSource"] = self.event_source
            if self.proxy_url:
                result["endpoints"]["proxyUrl"] = self.proxy_url

            if self.followers:
                result["followers"] = self.followers
            if self.following:
                result["following"] = self.following

        return result

    def _build_private_key(self):
        if self.public_key:
            return {
                "publicKey": {
                    "id": f"{self.account_url}#{self.public_key_name}",
                    "owner": self.account_url,
                    "publicKeyPem": self.public_key,
                }
            }
        return {}



def actor_for_object(data):
    if "attributedTo" in data:
        actor = data.get("attributedTo")
    else:
        actor = data.get("actor")
    actor = id_for_object(actor)

    if actor:
        return actor

    return "__NO__ACTOR__"


def id_for_object(data):
    if data is None:
        return None
    if isinstance(data, str):
        return data
    return data.get("id", None)


def property_for_key_as_set(data, key):
    if data is None:
        return set()
    result = data.get(key, [])
    if isinstance(result, str):
        return set([result])
    return set(result)


def recipients_for_object(data):
    return set.union(
        *[
            property_for_key_as_set(data, key)
            for key in ["to", "cc", "bto", "bcc", "audience"]
        ]
    )


def remove_public(recipients):
    return {
        x
        for x in recipients
        if x
        not in ["Public", "as:Public", "https://www.w3.org/ns/activitystreams#Public"]
    }


def is_public(data):
    recipients = recipients_for_object(data)

    return any(
        x in recipients
        for x in ["Public", "as:Public", "https://www.w3.org/ns/activitystreams#Public"]
    )



import json

from . import build_signature
from .crypto import content_digest_sha256, generate_public_private_key
from .http_signature import HttpSignature


def test_http_signature():
    public_key, private_key = generate_public_private_key()

    http_signature = HttpSignature().with_field("name", "value")

    signature_string = http_signature.build_signature("key_id", private_key)

    key_id, algorithm, headers, signature = signature_string.split(",")

    assert key_id == 'keyId="key_id"'
    assert algorithm == 'algorithm="rsa-sha256"'
    assert headers == 'headers="name"'
    assert signature.startswith("signature=")


def test_build_signature():
    public_key, private_key = generate_public_private_key()

    http_signature = build_signature("host", "method", "target")

    signature_string = http_signature.build_signature("key_id", private_key)

    key_id, algorithm, headers, signature = signature_string.split(",")

    assert key_id == 'keyId="key_id"'
    assert algorithm == 'algorithm="rsa-sha256"'
    assert headers == 'headers="(request-target) host"'
    assert signature.startswith("signature=")


def test_build_message():
    http_signature = build_signature(
        "myhost.tld", "get", "/path/to/resource"
    ).with_field("date", "Wed, 15 Mar 2023 17:28:15 GMT")

    message = http_signature.build_message()

    assert (
        message
        == """(request-target): get /path/to/resource
host: myhost.tld
date: Wed, 15 Mar 2023 17:28:15 GMT"""
    )

    signature = http_signature.ed25519_sign(
        "z3u2Yxcowsarethebestcowsarethebestcowsarethebest"
    )

    assert (
        signature
        == "z5ahdHCbP9aJEsDtvG1MEZpxPzuvGKYcdXdKvMq5YL21Z2"
        + "umxjs1SopCY2Ap8vZxVjTEf6dYbGuB7mtgcgUyNdBLe"
    )

    didkey = "did:key:z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5"

    http_signature.ed25519_verify(didkey, signature)


def test_build_message_post():
    body = json.dumps({"cows": "good"}).encode("utf-8")

    assert body == b'{"cows": "good"}'

    digest = content_digest_sha256(body)

    assert digest == "sha-256=MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k="

    http_signature = (
        build_signature("myhost.tld", "post", "/path/to/resource")
        .with_field("date", "Wed, 15 Mar 2023 17:28:15 GMT")
        .with_field("digest", digest)
    )

    message = http_signature.build_message()

    assert (
        message
        == """(request-target): post /path/to/resource
host: myhost.tld
date: Wed, 15 Mar 2023 17:28:15 GMT
digest: sha-256=MILb5lUDD6Z0pDSxhgxj+hMBEw0uTzP3g2qUJGHMp9k="""
    )

    signature = http_signature.ed25519_sign(
        "z3u2Yxcowsarethebestcowsarethebestcowsarethebest"
    )

    assert (
        signature
        == "z4vPkJaoaSVQp5DrMb8EvCajJcerW36rsyWDELTWQ3cYmaonnGf"
        + "b8WHiwH54BShidCcmpoyHjanVRYNrXXXka4jAn"
    )

    didkey = "did:key:z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5"

    http_signature.ed25519_verify(didkey, signature)



class UndoBuilder:
    def __init__(self, obj: dict):
        self.obj = obj

    def build(self) -> dict:
        return {
            "@context": "https://www.w3.org/ns/activitystreams",
            "type": "Undo",
            "actor": self.obj["actor"],
            "object": self.obj,
        }



import re

from bovine.activitystreams.objects import build_note

from .create_builder import CreateBuilder


def test_basic_build():
    result = CreateBuilder({"a": "message"}).build()

    assert "https://www.w3.org/ns/activitystreams" in result["@context"]
    assert result["type"] == "Create"
    assert result["id"] is None
    assert result["actor"] is None
    assert re.match(r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z", result["published"])


def test_basic_build_as_unlisted():
    result = (
        CreateBuilder({"a": "message"}).with_account("account").as_unlisted().build()
    )

    assert result["to"] == ["account/followers"]
    assert result["cc"] == ["https://www.w3.org/ns/activitystreams#Public"]


def test_copies_from_note_if_available():
    note = build_note("account", "url", "content").as_public().build()

    result = CreateBuilder(note).build()

    assert result["actor"] == "account"



from bovine.activitystreams.objects import build_note

from . import actor_for_object, is_public, recipients_for_object


def test_get_recipients():
    note = (
        build_note("account", "url", "content")
        .as_public()
        .add_cc("cc")
        .add_to("to")
        .add_cc("same")
        .add_to("same")
        .build()
    )

    recipients = recipients_for_object(note)

    assert recipients == {
        "account/followers",
        "https://www.w3.org/ns/activitystreams#Public",
        "cc",
        "to",
        "same",
    }


def test_is_public():
    note = build_note("account", "url", "content").as_public().build()
    assert is_public(note)

    note = build_note("account", "url", "content").build()
    assert not is_public(note)

    note = build_note("account", "url", "content").add_cc("someone").build()
    assert not is_public(note)

    note = build_note("account", "url", "content").as_unlisted().build()
    assert is_public(note)

    note = build_note("account", "url", "content").add_to("as:Public").build()
    note["to"] = "as:Public"
    assert is_public(note)


def test_actor_for_object():
    assert actor_for_object({}) == "__NO__ACTOR__"
    assert actor_for_object({"actor": "alice"}) == "alice"
    assert actor_for_object({"actor": {"id": "alice"}}) == "alice"
    assert actor_for_object({"actor": {}}) == "__NO__ACTOR__"
    assert actor_for_object({"attributedTo": "alice"}) == "alice"
    assert actor_for_object({"attributedTo": {"id": "alice"}}) == "alice"
    assert actor_for_object({"attributedTo": {}}) == "__NO__ACTOR__"



import uuid

from bovine.activitystreams.common import build_context


class LikeBuilder:
    def __init__(self, account: str, obj: dict):
        self.account = account
        self.obj = obj
        self.content = None
        self.to = set()
        self.cc = set()

    def add_to(self, recipient):
        self.to.add(recipient)
        return self

    def add_cc(self, recipient):
        self.cc.add(recipient)
        return self

    def with_content(self, content):
        self.content = content
        return self

    def build(self) -> dict:
        context = build_context().build()

        result = {
            "@context": context,
            "id": self.account + "#likes+" + str(uuid.uuid4()),
            "type": "Like",
            "actor": self.account,
            "object": self.obj,
        }

        if self.to:
            result["to"] = list(self.to)

        if self.cc:
            result["cc"] = list(self.cc)

        if self.content:
            result["content"] = self.content

        return result



from datetime import datetime


class CreateBuilder:
    def __init__(self, obj: dict):
        self.obj = obj
        self.account = self.obj.get("attributedTo", None)
        self.url = self.obj.get("id", None)
        self.to = self.obj.get("to", [])
        self.cc = self.obj.get("cc", [])
        self.published = self.obj.get(
            "published", datetime.utcnow().replace(microsecond=0).isoformat() + "Z"
        )

    def with_account(self, account: str):
        self.account = account
        return self

    def as_public(self):
        self.to.append("https://www.w3.org/ns/activitystreams#Public")
        self.cc.append(f"{self.account}/followers")
        return self

    def as_unlisted(self):
        self.to.append(f"{self.account}/followers")
        self.cc.append("https://www.w3.org/ns/activitystreams#Public")
        return self

    def build(self) -> dict:
        return {
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                {
                    "inReplyToAtomUri": "ostatus:inReplyToAtomUri",
                    "conversation": "ostatus:conversation",
                    "ostatus": "http://ostatus.org#",
                },
            ],
            "id": self.url,
            "type": "Create",
            "actor": self.account,
            "object": self.obj,
            "published": self.published,
            "to": self.to,
            "cc": self.cc,
        }



from .like_builder import LikeBuilder


def test_like_builder():
    result = (
        LikeBuilder("account", "obj")
        .add_to("target")
        .add_cc("other")
        .with_content("🐮")
        .build()
    )

    assert result["@context"] == "https://www.w3.org/ns/activitystreams"

    assert result["to"] == ["target"]
    assert result["cc"] == ["other"]

    assert result["type"] == "Like"



from bovine.types import Visibility

from .actor_builder import ActorBuilder


def test_basic_build():
    result = ActorBuilder("name").build()

    assert result["@context"] == "https://www.w3.org/ns/activitystreams"
    assert result["name"] == "name"
    assert result["type"] == "Person"


def test_account_urls():
    result = ActorBuilder("name").with_account_url("account_url").build()

    assert result["id"] == "account_url"
    assert result["inbox"] == "account_url/inbox"
    assert result["outbox"] == "account_url/outbox"


def test_public_key():
    result = (
        ActorBuilder("name")
        .with_account_url("account_url")
        .with_public_key("---key---")
        .build()
    )

    assert isinstance(result["@context"], list)
    assert "https://www.w3.org/ns/activitystreams" in result["@context"]
    assert "https://w3id.org/security/v1" in result["@context"]

    assert result["publicKey"] == {
        "id": "account_url#main-key",
        "owner": "account_url",
        "publicKeyPem": "---key---",
    }


def test_visibility_web():
    result = (
        ActorBuilder("name")
        .with_account_url("account_url")
        .with_public_key("---key---")
        .build(visibility=Visibility.WEB)
    )

    assert isinstance(result["@context"], list)
    assert "https://www.w3.org/ns/activitystreams" in result["@context"]
    assert "https://w3id.org/security/v1" in result["@context"]

    assert "inbox" not in result

    assert result["publicKey"] == {
        "id": "account_url#main-key",
        "owner": "account_url",
        "publicKeyPem": "---key---",
    }



from .object_builder import ObjectBuilder


class NoteBuilder(ObjectBuilder):
    def __init__(self, account, content, followers=None):
        super().__init__("Note", account, followers=followers)

        self.with_content(content)



from .note_builder import NoteBuilder


def test_note_builder_basic():
    result = NoteBuilder("account", "content").build()

    assert result["@context"] == "https://www.w3.org/ns/activitystreams"
    assert result["type"] == "Note"
    assert result["content"] == "content"


def test_note_builder_hashtags():
    result = NoteBuilder("account", "content").with_hashtag("#tag1").build()

    assert isinstance(result["tag"], list)
    assert result["tag"][0] == {"name": "#tag1", "type": "Hashtag"}


def test_note_builder_hashtag_and_mention():
    result = (
        NoteBuilder("account", "content")
        .with_hashtag("#tag1")
        .with_mention("proto://server/path/mention")
        .build()
    )

    assert isinstance(result["tag"], list)
    assert result["tag"][0] == {"name": "#tag1", "type": "Hashtag"}
    assert result["tag"][1] == {
        "href": "proto://server/path/mention",
        "name": "proto://server/path/mention",
        "type": "Mention",
    }
    assert result["cc"] == ["proto://server/path/mention"]


def test_note_builder_cc():
    result = (
        NoteBuilder("account", "content", followers="account/followers")
        .as_public()
        .add_cc("user")
        .build()
    )

    assert set(result["cc"]) == {"account/followers", "user"}


def test_note_builder_cc_and_to():
    result = (
        NoteBuilder("account", "content", followers="account/followers")
        .as_public()
        .add_cc("user")
        .add_cc("user")
        .add_cc("to_user")
        .add_to("to_user")
        .build()
    )

    assert set(result["cc"]) == {"account/followers", "user"}
    assert "to_user" in result["to"]


def test_note_builder_for_reply():
    result = (
        NoteBuilder("account", "content")
        .with_conversation("conversation")
        .with_reply("reply_id")
        .build()
    )

    assert result["conversation"] == "conversation"
    assert result["inReplyTo"] == "reply_id"


def test_note_builder_for_source():
    result = NoteBuilder("account", "content").with_source("source", "format").build()

    assert result["source"] == {"content": "source", "mediaType": "format"}



import base64
import hashlib
import logging

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    load_pem_public_key,
)

logger = logging.getLogger(__name__)


def generate_public_private_key():
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
    private_key_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    public_key = private_key.public_key()
    public_key_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )

    return public_key_pem.decode("utf-8"), private_key_pem.decode("utf-8")


def sign_message(private_key, message):
    key = load_pem_private_key(private_key.encode("utf-8"), password=None)

    return base64.standard_b64encode(
        key.sign(
            message.encode("utf-8"),
            padding.PKCS1v15(),
            hashes.SHA256(),
        )
    ).decode("utf-8")


def verify_signature(public_key, message, signature):
    public_key_loaded = load_pem_public_key(public_key.encode("utf-8"))

    try:
        public_key_loaded.verify(
            base64.standard_b64decode(signature),
            message.encode("utf-8"),
            padding.PKCS1v15(),
            hashes.SHA256(),
        )
    except InvalidSignature:
        logger.warning("invalid signature")
        return False

    return True


def content_digest_sha256(content):
    if isinstance(content, str):
        content = content.encode("utf-8")

    digest = base64.standard_b64encode(hashlib.sha256(content).digest()).decode("utf-8")
    return "sha-256=" + digest



from .follow_builder import FollowBuilder


def test_follow_builder() -> None:
    data = FollowBuilder("domain", "actor", "tofollow").build()

    assert data["@context"] == "https://www.w3.org/ns/activitystreams"
    assert data["id"].startswith("https://domain/")
    assert data["type"] == "Follow"
    assert data["actor"] == "actor"
    assert data["object"] == "tofollow"



from bovine.activitystreams.common import as_public, build_context
from bovine.activitystreams.objects import tombstone


class DeleteBuilder:
    def __init__(self, actor, object_id):
        self.actor = actor
        self.object_id = object_id
        self.context_builder = build_context()

    def build(self):
        my_tombstone = tombstone(self.object_id)
        my_tombstone = self.context_builder.absorb(my_tombstone)

        return {
            "@context": self.context_builder.build(),
            "id": self.object_id + "/delete",
            "type": "Delete",
            "actor": self.actor,
            "to": [as_public],
            "cc": [self.actor + "/followers"],
            "object": my_tombstone,
        }



import json
from unittest.mock import AsyncMock, MagicMock

import aiohttp

from bovine.activitystreams import (
    build_ordered_collection,
    build_ordered_collection_page,
)
from bovine.utils.test import get_user_keys

from .actor import ActivityPubActor


def test_actor_host():
    actor = ActivityPubActor().with_actor_id("https://domain.tld/users/someone")

    assert actor.host == "domain.tld"


async def test_activity_pub_client_get_collection_no_pages():
    session = AsyncMock(aiohttp.ClientSession)
    url = "https://test_domain/test_path"
    public_key_url = "public_key_url"
    public_key, private_key = get_user_keys()
    session = AsyncMock(aiohttp.ClientSession)
    session.get = AsyncMock()

    actor = (
        ActivityPubActor()
        .with_actor_id("actor_id")
        .with_http_signature(public_key_url, private_key, session=session)
    )

    text_mock = AsyncMock()
    session.get.return_value = MagicMock(aiohttp.ClientResponse)
    session.get.return_value.text = text_mock

    items = [{"id": j} for j in range(7)]

    builder = build_ordered_collection("url").with_count(7).with_items(items)

    text_mock.return_value = json.dumps(builder.build())

    result = await actor.get_ordered_collection(url)
    session.get.assert_awaited_once()

    assert result["total_items"] == 7
    assert result["items"] == items


async def test_activity_pub_client_get_collection_pages():
    session = AsyncMock(aiohttp.ClientSession)
    url = "https://test_domain/test_path"
    public_key_url = "public_key_url"
    public_key, private_key = get_user_keys()
    session = AsyncMock(aiohttp.ClientSession)
    session.get = AsyncMock()

    actor = (
        ActivityPubActor()
        .with_actor_id("actor_id")
        .with_http_signature(public_key_url, private_key, session=session)
    )
    text_mock = AsyncMock()
    session.get.return_value = MagicMock(aiohttp.ClientResponse)
    session.get.return_value.text = text_mock

    items = [{"id": j} for j in range(23)]

    builder = (
        build_ordered_collection("url")
        .with_count(23)
        .with_first_and_last("first", "last")
    )

    page_1 = (
        build_ordered_collection_page("url_1", "page_1")
        .with_items(items[:13])
        .with_next("next_1")
    )
    page_2 = (
        build_ordered_collection_page("url_1", "page_2")
        .with_items(items[13:20])
        .with_next("next_2")
    )
    page_3 = build_ordered_collection_page("url_1", "page_2").with_items(items[20:])

    text_mock.side_effect = [
        json.dumps(builder.build()),
        json.dumps(page_1.build()),
        json.dumps(page_2.build()),
        json.dumps(page_3.build()),
    ]

    result = await actor.get_ordered_collection(url)

    assert result["total_items"] == 23
    assert result["items"] == items

    text_mock.side_effect = [
        json.dumps(builder.build()),
        json.dumps(page_1.build()),
        json.dumps(page_2.build()),
        json.dumps(page_3.build()),
    ]

    result = await actor.get_ordered_collection(url, max_items=17)

    assert result["total_items"] == 23
    assert result["items"] == items[:20]



from .delete_builder import DeleteBuilder


def test_delete_builder():
    result = DeleteBuilder("actor", "object_id").build()

    assert result["id"] == "object_id/delete"
    assert result["type"] == "Delete"
    assert "@context" not in result["object"]



import json
import logging

import aiohttp

from bovine.utils.parse import parse_fediverse_handle

from .consts import BOVINE_CLIENT_NAME

logger = logging.getLogger(__name__)


async def lookup_with_webfinger(
    session: aiohttp.ClientSession, webfinger_url: str, params: dict
):
    async with session.get(
        webfinger_url, params=params, headers={"user-agent": BOVINE_CLIENT_NAME}
    ) as response:
        if response.status != 200:
            logger.warn(f"{params['resource']} not found using webfinger")
            return None
        text = await response.text()
        data = json.loads(text)

        if "links" not in data:
            return None

        links = data["links"]
        for entry in links:
            if "rel" in entry and entry["rel"] == "self":
                return entry["href"]

    return None


async def lookup_account_with_webfinger(
    session: aiohttp.ClientSession, fediverse_handle: str
) -> str | None:
    username, domain = parse_fediverse_handle(fediverse_handle)

    webfinger_url = f"https://{domain}/.well-known/webfinger"
    params = {"resource": f"acct:{username}@{domain}"}

    return await lookup_with_webfinger(session, webfinger_url, params)


async def lookup_did_with_webfinger(
    session: aiohttp.ClientSession, domain: str, did: str
) -> str | None:
    webfinger_url = f"https://{domain}/.well-known/webfinger"
    params = {"resource": did}

    return await lookup_with_webfinger(session, webfinger_url, params)



from .accept_builder import AcceptBuilder


def test_accept_builder():
    result = AcceptBuilder("account", {"an": "object"}).build()

    assert result["@context"] == "https://www.w3.org/ns/activitystreams"
    assert result["id"] == "account#accepts/follows/"
    assert result["type"] == "Accept"
    assert result["actor"] == "account"
    assert result["object"]["an"] == "object"



import json
import logging
from urllib.parse import urlparse

import aiohttp
import tomli

from bovine.clients.lookup_account import lookup_did_with_webfinger
from bovine.clients.moo_auth_client import MooAuthClient
from bovine.clients.signed_http_client import SignedHttpClient
from bovine.utils.crypto.did_key import private_key_to_did_key

from .activity_factory import ActivityFactory
from .collection_helper import CollectionHelper
from .object_factory import ObjectFactory

logger = logging.getLogger(__name__)


class ActivityPubActor:
    def __init__(self):
        self.actor_id = None
        self.client = None
        self.information = None
        self._activity_factory = None
        self._object_factory = None

    async def with_host_and_ed25519_private_key(self, host, private_key, session=None):
        if session is None:
            session = aiohttp.ClientSession()
        did_key = private_key_to_did_key(private_key)

        self.actor_id = await lookup_did_with_webfinger(session, host, did_key)

        self.client = MooAuthClient(session, did_key, private_key)

        # await self.load()

        return self

    def with_actor_id(self, actor_id):
        self.actor_id = actor_id
        return self

    def with_http_signature(self, public_key_url, private_key, session=None):
        if session is None:
            session = aiohttp.ClientSession()

        self.client = SignedHttpClient(session, public_key_url, private_key)

        return self

    async def load(self):
        if self.client is None:
            raise Exception("Client not set in ActivityPubActor")
        self.information = await self.get(self.actor_id)

        logger.debug("Retrieved information %s", self.information)

        if any(required not in self.information for required in ["inbox", "outbox"]):
            raise Exception("Retrieved incomplete actor data")

    async def send_to_outbox(self, data: dict):
        if self.information is None:
            await self.load()

        return await self.post(self.information["outbox"], data)

    async def post(self, target, data: dict):
        response = await self.client.post(target, json.dumps(data))
        response.raise_for_status()

        return response

    async def proxy_element(self, target):
        response = await self.client.post(
            self.information["endpoints"]["proxyUrl"],
            f"id={target}",
            content_type="application/x-www-form-urlencoded",
        )
        response.raise_for_status()
        return await response.json()

    async def get_ordered_collection(self, url, max_items=None):
        result = await self.client.get(url)
        result.raise_for_status()

        data = json.loads(await result.text())

        total_number_of_items = data["totalItems"]
        items = []

        if "orderedItems" in data:
            items = data["orderedItems"]

        if len(items) == total_number_of_items:
            return {"total_items": total_number_of_items, "items": items}

        if "first" in data:
            page_data = await self.get(data["first"])

            items = page_data["orderedItems"]

            while "next" in page_data and len(page_data["orderedItems"]) > 0:
                if max_items and len(items) > max_items:
                    return {"total_items": total_number_of_items, "items": items}

                page_data = await self.get(page_data["next"])

                items += page_data["orderedItems"]

        return {"total_items": total_number_of_items, "items": items}

    async def get(self, target):
        response = await self.client.get(target)
        response.raise_for_status()
        return json.loads(await response.text())

    async def event_source(self):
        if self.information is None:
            await self.load()

        event_source_url = self.information["endpoints"]["eventSource"]
        return self.client.event_source(event_source_url)

    @property
    def activity_factory(self):
        if self._activity_factory is None:
            self._activity_factory = ActivityFactory(self.information)
        return self._activity_factory

    @property
    def object_factory(self):
        if self._object_factory is None:
            self._object_factory = ObjectFactory(self.information)
        return self._object_factory

    @property
    def factories(self):
        return self.activity_factory, self.object_factory

    @property
    def host(self):
        return urlparse(self.actor_id).netloc

    async def inbox(self):
        inbox_collection = CollectionHelper(self.information["inbox"], self)
        await inbox_collection.refresh()
        return inbox_collection

    async def outbox(self):
        inbox_collection = CollectionHelper(self.information["outbox"], self)
        await inbox_collection.refresh()
        return inbox_collection

    @staticmethod
    def from_file(filename, session):
        with open(filename, "rb") as fp:
            data = tomli.load(fp)

        actor = ActivityPubActor(data["account_url"])
        actor.with_http_signature(
            data["public_key_url"], data["private_key"], session=session
        )

        return actor



import aiohttp

from .lookup_account import lookup_account_with_webfinger, lookup_did_with_webfinger


async def test_lookup_account():
    async with aiohttp.ClientSession() as session:
        result = await lookup_account_with_webfinger(session, "helge@mymath.rocks")

    assert result.startswith("https://mymath.rocks/")


async def test_lookup_did_with_webfinger():
    did_helge = "did:key:z6MkujdZ216eYz55vz8X5HetqeJXj9ddn5ZHZUsBpRX4wfnL"
    async with aiohttp.ClientSession() as session:
        result = await lookup_did_with_webfinger(session, "mymath.rocks", did_helge)

    assert result.startswith("https://mymath.rocks/")



from .common import build_context


class OrderedCollectionBuilder:
    def __init__(self, url: str):
        self.context_builder = build_context()
        self.url = url
        self.items: list | None = None
        self.count = 0
        self.first = None
        self.last = None

    def with_items(self, items: list):
        self.items = items
        return self

    def with_count(self, count: int):
        self.count = count
        return self

    def with_first_and_last(self, first, last):
        self.first = first
        self.last = last
        return self

    def build(self) -> dict:
        result = {
            "@context": self.context_builder.build(),
            "id": self.url,
            "totalItems": self.count,
            "type": "OrderedCollection",
        }

        if self.items:
            result["orderedItems"] = self.items

        if self.first:
            result["first"] = self.first

        if self.last:
            result["last"] = self.last

        return result



from bovine.utils.test import get_user_keys

from . import (
    content_digest_sha256,
    generate_public_private_key,
    sign_message,
    verify_signature,
)


def test_crypto_sign_verify():
    message = "secret"

    public_key, private_key = get_user_keys()

    signature = sign_message(private_key, message)

    assert verify_signature(public_key, message, signature)


def test_crypto_sign_verify_failure():
    message = "secret"

    public_key, private_key = get_user_keys()

    assert not verify_signature(public_key, message, "")


def test_content_digest_sha256():
    digest = content_digest_sha256("content")

    assert digest.startswith("sha-256=")


def test_generate_public_private_key():
    public_key, private_key = generate_public_private_key()

    assert public_key.startswith("-----BEGIN PUBLIC KEY-----")
    assert public_key.endswith("-----END PUBLIC KEY-----\n")

    assert private_key.startswith("-----BEGIN PRIVATE KEY-----")
    assert private_key.endswith("-----END PRIVATE KEY-----\n")



from cryptography.hazmat.primitives.asymmetric import ed25519

from .did_key import (
    did_key_to_public_key,
    generate_keys,
    private_key_to_did_key,
    public_key_to_did_key,
)

did_example = "did:key:z6MkiTBz1ymuepAQ4HEHYSF1H8quG5GLVVQR3djdX3mDooWp"


def test_encode_private_key():
    _, public_key = generate_keys()
    did = public_key_to_did_key(public_key)

    # See https://w3c-ccg.github.io/did-method-key/#ed25519-x25519
    # These DID always start with z6Mk.

    assert did.startswith("did:key:z6Mk")
    assert len(did) == len(did_example)


def test_did_to_public_key():
    public_key = did_key_to_public_key(did_example)

    assert isinstance(public_key, ed25519.Ed25519PublicKey)


def test_encode_private_key_and_back():
    private_key, public_key = generate_keys()
    did = public_key_to_did_key(public_key)
    transformed = did_key_to_public_key(did)

    message = b"Hello did-core!"

    signature = private_key.sign(message)

    # If verification fails an error is thrown

    public_key.verify(signature, message)
    transformed.verify(signature, message)


def test_private_key_to_did_key():
    private_key = "z3u2Yxcowsarethebestcowsarethebestcowsarethebest"

    did_key = private_key_to_did_key(private_key)

    assert did_key == "did:key:z6MkekwC6R9bj9ErToB7AiZJfyCSDhaZe1UxhDbCqJrhqpS5"



from datetime import datetime
from typing import Set


class ObjectBuilder:
    def __init__(self, object_type: str, account: str, followers=None):
        self.type = object_type
        self.account = account
        self.content = None
        self.name = None
        self.summary = None
        self.url = None
        self.to: Set[str] = set()
        self.cc: Set[str] = set()
        self.published = datetime.utcnow().replace(microsecond=0).isoformat() + "Z"
        self.hashtags: Set[str] = set()
        self.mentions: Set[str] = set()
        self.conversation: str | None = None
        self.reply_to_id: str | None = None
        self.source: dict | None = None
        self.followers = followers

    def with_content(self, content: str):
        self.content = content
        return self

    def with_url(self, url: str):
        self.url = url
        return self

    def with_name(self, name: str):
        self.name = name
        return self

    def with_summary(self, summary: str):
        self.summary = summary
        return self

    def as_public(self, followers=None):
        # FIXME: Broken in the way followers is added
        self.to.add("https://www.w3.org/ns/activitystreams#Public")
        if followers:
            self.cc.add(followers)
        elif self.followers:
            self.cc.add(self.followers)
        else:
            raise Exception("Followers unkown")
        return self

    def add_cc(self, recipient: str):
        self.cc.add(recipient)
        return self

    def add_to(self, recipient: str):
        self.to.add(recipient)
        return self

    def as_unlisted(self, followers=None):
        if followers:
            self.cc.add(followers)
        elif self.followers:
            self.cc.add(self.followers)
        else:
            raise Exception("Followers unkown")
        self.cc.add("https://www.w3.org/ns/activitystreams#Public")
        return self

    def with_hashtag(self, hashtag: str):
        self.hashtags.add(hashtag)
        return self

    def with_mention(self, mention: str):
        self.mentions.add(mention)
        self.cc.add(mention)
        return self

    def with_conversation(self, conversation: str):
        self.conversation = conversation
        return self

    def with_reply(self, rid: str):
        self.reply_to_id = rid
        return self

    def with_source(self, content: str, media_type: str):
        self.source = {"content": content, "mediaType": media_type}
        return self

    def build(self) -> dict:
        result: dict = {
            "@context": "https://www.w3.org/ns/activitystreams",
            "attributedTo": self.account,
            "type": self.type,
            "inReplyTo": None,
            "content": self.content,
            "published": self.published,
            "to": list(self.to),
            "cc": list(self.cc - self.to),
        }

        result = self._add_tag(result)

        extra_fields = {
            "inReplyTo": self.reply_to_id,
            "conversation": self.conversation,
            "source": self.source,
            "name": self.name,
            "url": self.url,
            "summary": self.summary,
        }

        for key, value in extra_fields.items():
            if value:
                result[key] = value

        return result

    def _add_tag(self, result):
        tag_list = [{"name": tag, "type": "Hashtag"} for tag in self.hashtags] + [
            {"href": mention, "name": mention, "type": "Mention"}
            for mention in self.mentions
        ]

        if len(tag_list) > 0:
            result["tag"] = tag_list

        return result

    def with_published(self, published: datetime):
        self.published = published.replace(microsecond=0).isoformat() + "Z"
        return self



from textwrap import wrap

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519
from multiformats import multibase, multicodec


def generate_keys():
    private_key = ed25519.Ed25519PrivateKey.generate()
    public_key = private_key.public_key()

    return private_key, public_key


def encode_private_key(private_key):
    private_bytes = private_key.private_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PrivateFormat.Raw,
        encryption_algorithm=serialization.NoEncryption(),
    )

    wrapped = multicodec.wrap("ed25519-priv", private_bytes)

    return multibase.encode(wrapped, "base58btc")


def encode_public_key(public_key):
    public_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw,
    )

    wrapped = multicodec.wrap("ed25519-pub", public_bytes)
    encoded = multibase.encode(wrapped, "base58btc")

    return encoded


def public_key_to_did_key(public_key):
    return "did:key:" + encode_public_key(public_key)


def private_key_to_secret(private_key):
    return "secret_:" + encode_private_key(private_key)


def public_key_to_hey_key(public_key):
    encoded = encode_public_key(public_key)
    return "hey:key:" + " ".join(wrap(encoded, 4))


def private_key_to_hey_secret(private_key):
    encoded = encode_private_key(private_key)
    return "hey:sec:" + " ".join(wrap(encoded, 4))


def did_key_to_public_key(did):
    assert did.startswith("did:key:")
    decoded = multibase.decode(did[8:])
    codec, key_bytes = multicodec.unwrap(decoded)
    assert codec.name == "ed25519-pub"

    return ed25519.Ed25519PublicKey.from_public_bytes(key_bytes)


def private_key_to_ed25519(private_key_str):
    decoded = multibase.decode(private_key_str)
    codec, key_bytes = multicodec.unwrap(decoded)
    assert codec.name == "ed25519-priv"

    return ed25519.Ed25519PrivateKey.from_private_bytes(key_bytes)


def private_key_to_did_key(private_key_str):
    private_key = private_key_to_ed25519(private_key_str)

    return public_key_to_did_key(private_key.public_key())



import uuid


class FollowBuilder:
    def __init__(self, domain: str, actor: str, tofollow: str):
        self.domain = domain
        self.actor = actor
        self.tofollow = tofollow

    def build(self) -> dict:
        uuid_string = str(uuid.uuid4())

        return {
            "@context": "https://www.w3.org/ns/activitystreams",
            "id": f"https://{self.domain}/{uuid_string}",
            "type": "Follow",
            "actor": self.actor,
            "object": self.tofollow,
        }



from bovine.activitystreams.common import build_context


class TombstoneBuilder:
    def __init__(self, object_id):
        self.object_id = object_id

    def build(self):
        context = build_context().build()

        return {
            "@context": context,
            "type": "Tombstone",
            "id": self.object_id,
        }



import aiohttp
import tomli

from .activitypub.actor import ActivityPubActor


class BovineActor(ActivityPubActor):
    def __init__(self, config):
        self.config = config
        super().__init__()

    async def init(self, session=None):
        self.session = session
        if session is None:
            self.session = aiohttp.ClientSession()

        if self._has_moo_auth():
            await self.with_host_and_ed25519_private_key(
                self.config["host"], self.config["private_key"], session=self.session
            )
        elif self._has_http_signature():
            self.actor_id = self.config["account_url"]
            self.with_http_signature(
                self.config["public_key_url"],
                self.config["private_key"],
                session=self.session,
            )
        else:
            raise Exception("No known authorization method available")

        await self.load()

    async def __aenter__(self):
        await self.init()
        return self

    async def __aexit__(self, *args):
        await self.session.close()

    def _has_http_signature(self):
        return self._has_keys(["account_url", "public_key_url", "private_key"])

    def _has_moo_auth(self):
        return self._has_keys(["host", "private_key"])

    def _has_keys(self, key_list):
        for key in key_list:
            if key not in self.config:
                return False

        return True

    @staticmethod
    def from_file(config_file):
        with open(config_file, "rb") as fp:
            config = tomli.load(fp)

        return BovineActor(config)



from .ordered_collection_builder import OrderedCollectionBuilder


def test_ordered_collection_builder():
    result = OrderedCollectionBuilder("url").build()

    assert result == {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "url",
        "totalItems": 0,
        "type": "OrderedCollection",
    }


def test_ordered_collection_builder_with_items():
    result = (
        OrderedCollectionBuilder("url")
        .with_count(1)
        .with_items([{"item": "1"}])
        .build()
    )

    assert result == {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "url",
        "orderedItems": [{"item": "1"}],
        "totalItems": 1,
        "type": "OrderedCollection",
    }


def test_ordered_collection_builder_with_fist_last():
    result = (
        OrderedCollectionBuilder("url")
        .with_count(1)
        .with_first_and_last("first", "last")
        .build()
    )

    assert result == {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "url",
        "first": "first",
        "last": "last",
        "totalItems": 1,
        "type": "OrderedCollection",
    }



from .ordered_collection_page_builder import OrderedCollectionPageBuilder


def test_ordered_collection_page_builder():
    result = (
        OrderedCollectionPageBuilder("url?page=1", "url").with_items(["id1"]).build()
    )

    assert result == {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "url?page=1",
        "partOf": "url",
        "orderedItems": ["id1"],
        "type": "OrderedCollectionPage",
    }

    result = (
        OrderedCollectionPageBuilder("url?page=1", "url")
        .with_items(["id1"])
        .with_next("next_url")
        .with_prev("prev_url")
        .build()
    )

    assert result == {
        "@context": "https://www.w3.org/ns/activitystreams",
        "id": "url?page=1",
        "next": "next_url",
        "prev": "prev_url",
        "partOf": "url",
        "orderedItems": ["id1"],
        "type": "OrderedCollectionPage",
    }



from bovine.activitystreams.activities import build_accept, build_create, build_like


class ActivityFactory:
    def __init__(self, actor_information):
        self.information = actor_information

    def create(self, note):
        return build_create(note)

    def like(self, target):
        return build_like(self.information["id"], target)

    def accept(self, obj):
        return build_accept(self.information["id"], obj)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bovine',
 'bovine.activitypub',
 'bovine.activitystreams',
 'bovine.activitystreams.activities',
 'bovine.activitystreams.common',
 'bovine.activitystreams.objects',
 'bovine.activitystreams.utils',
 'bovine.clients',
 'bovine.utils',
 'bovine.utils.crypto']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.3,<4.0.0',
 'bleach>=6.0.0,<7.0.0',
 'cryptography>=39.0.0,<40.0.0',
 'multiformats>=0.2.1,<0.3.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'tomli>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'bovine',
    'version': '0.0.9',
    'description': 'Core functionality of bovine needed to build client to server applications',
    'long_description': '# Bovine\n\nThis package contains two essential parts of bovine. First\nit defines `BovineActor`, which contains all the necessities\nto write ActivityPub Clients. Furthermore, this package contains\nthe cryptographic routines to verify HTTP signatures.\n\nFurthermore, the folder `examples` contains a few examples on\nhow `BovineActor` can be used. The cryptographic routines\nare used in `bovine_fedi` to verify signatures.\n\n## Example: Make a post aka Faking at being a Server\n\nWhile [ActivityPub](https://www.w3.org/TR/activitypub/) specifies Server to Server and Client to Server, they really are just two sides of the same coin. In this example, we will work through how to use `BovineActor` to post a message.\n\nWithout having an ActivityPub Server supporting Client to Server, this will require a bit of setup. This setup will build a stub server that just allows other ActivityPub servers to associate us with a domain.\n\nThe stub server is given by the following snippet. One should note that it just answers with predefined json from a config file, that hasn\'t been generated yet. One could easily replace it with serving static files. See also [the Mastodon Blog](https://blog.joinmastodon.org/2018/06/how-to-implement-a-basic-activitypub-server/) for a similar implementation.\n\n```python\nimport tomli\nimport json\nfrom quart import Quart\n\napp = Quart(__name__)\n\nwith open("server.toml", "rb") as fp:\n    config = tomli.load(fp)\n\n@app.get("/.well-known/webfinger")\nasync def webfinger():\n    return json.loads(config["webfinger"])\n\n\n@app.get("/actor")\nasync def actor():\n    return json.loads(config["actor"])\n\nif __name__ == "__main__":\n    app.run()\n```\n\nThe following script generates the config files. You will have to adapt the `hostname` variable and be able to serve the entire thing through https.\n\n```python\nimport bovine\nimport tomli_w\nimport json\n\nhostname = "bovine-demo.mymath.rocks"\n\npublic_key, private_key = bovine.utils.crypto.generate_public_private_key()\nactor_url = f"https://{hostname}/actor"\nactor = (\n    bovine.activitystreams.build_actor("actor")\n    .with_account_url(actor_url)\n    .with_public_key(public_key)\n)\n\nwebfinger = {\n    "subject": f"acct:actor@{hostname}",\n    "links": [\n        {\n            "href": actor_url,\n            "rel": "self",\n            "type": "application/activity+json",\n        }\n    ],\n}\n\nserver_config = {"actor": json.dumps(actor.build()), "webfinger": json.dumps(webfinger)}\n\nactor_config = {\n    "account_url": actor_url,\n    "public_key_url": f"{actor_url}#main-key",\n    "private_key": private_key,\n}\n\nwith open("server.toml", "wb") as fp:\n    tomli_w.dump(server_config, fp)\n\nwith open("bovine.toml", "wb") as fp:\n    tomli_w.dump(actor_config, fp)\n```\n\nYou can now access the urls, which are in my case [https://bovine-demo.mymath.rocks/actor](https://bovine-demo.mymath.rocks/actor) and [https://bovine-demo.mymath.rocks/.well-known/webfinger?resource=acct:actor@bovine-demo.mymath.rocks](https://bovine-demo.mymath.rocks/.well-known/webfinger?resource=acct:actor@bovine-demo.mymath.rocks). Using this, we can now lookup the fediverse handle `actor@bovine-demo.mymath.rocks` on most FediVerse applications.\n\nYou can now send a post via the following code snippet:\n\n```python\nimport asyncio\n\nfrom uuid import uuid4\nfrom bovine import BovineActor\n\ntarget_account = "https://mas.to/users/themilkman"\ntarget_inbox = "https://mas.to/users/themilkman/inbox"\n\n\nasync def run():\n    async with BovineActor.from_file("bovine.toml") as actor:\n        activity_factory, object_factory = actor.factories\n        note = (\n            object_factory.note("Hello")\n            .add_to(target_account)\n            .with_mention(target_account)\n            .build()\n        )\n        note["id"] = actor.actor_id + "/" + str(uuid4())\n        create = activity_factory.create(note).build()\n        create["id"] = actor.actor_id + "/" + str(uuid4())\n\n        await actor.post(target_inbox, create)\n\n\nasyncio.run(run())\n```\n\nA few comments are in order:\n\n- The id needs to be set on the Note and Create in order to be compatible with Mastodon. When using proper Client To Server as below, it is superfluous\n- The form of adding the `target_account` to both to and mention causes it to be a direct message.\n- Normally, you would look up the inbox from the target account.\n\n## Using BovineActor\n\nOne can import it via `from bovine import BovineActor`. Then one can either use it via:\n\n```python\nasync with BovineActor(config) as actor:\n    ...\n# or\nactor = BovineActor(config)\nawait actor.init()\n```\n\nHere the config object can be present in two variants. First it can contain the keys `host` and `private_key`, where `host` is the domain the ActivityPub Actor is on and `private_key` is a mutlicodec encoded Ed25519 key, whose corresponding did-key has been added to the Actor. In this case [Moo-Auth-1](https://blog.mymath.rocks/2023-03-15/BIN1_Moo_Authentication_and_Authoriation) will be used. The second variant is to use [HTTP Signatures](https://docs.joinmastodon.org/spec/security/#http), where the keys `account_url`, `public_key_url`, and `private_key` need to be present. Alternatively, to passing a config object, one can use `BovineActor.from_file(path_to_toml_file)`.\n\n### Making a post\n\nBovineActor contains two factories to create [ActivityStreams Objects](https://www.w3.org/TR/activitystreams-vocabulary/#object-types) and [ActivityStreams Activities](https://www.w3.org/TR/activitystreams-vocabulary/#activity-types). One can obtain them by running\n\n```python\nactivity_factory, object_factory = actor.factories\n```\n\nThe simplest usage example is a create wrapping a note, that looks like:\n\n```python\nactivity_factory, object_factory = actor.factories\nnote = object_factory.note("Hello").as_public().build()\ncreate = activity_factory.create(note).build()\n```\n\nThe result should be the something equivalent to the json\n\n```json\n{\n  "@context": "https://www.w3.org/ns/activitystreams",\n  "type": "Create",\n  "actor": "https://domain/actor",\n  "object": {\n    "attributedTo": "https://domain/actor",\n    "type": "Note",\n    "content": "Hello",\n    "published": "2023-03-25T08:12:32Z",\n    "to": "as:Public",\n    "cc": "https://domain/followers_collection"\n  },\n  "published": "2023-03-25T08:12:32Z",\n  "to": "as:Public",\n  "cc": "https://domain/followers_collection"\n}\n```\n\nThe details depend on the used actor and will likely contain superfluous elements until the creation process is improved. We can now send this activity to our outbox using\n\n```python\nawait actor.send_to_outbox(create)\n```\n\n__Note__: This is different from what we did in the first example, where we used `await actor.post(inbox, create)`. The difference is that in the first example, we faked being a server, now we are actually using Client To Server.\n\n### The inbox and outbox\n\nBy running\n\n```python\ninbox = await actor.inbox()\noutbox = await actor.outbox()\n```\n\none can obtain `CollectionHelper` objects. These are meant to make it easier to interact with collection objects. In the simplest use case, one can use\n\n```python\nawait inbox.next_item()\n```\n\nto get the items from the inbox one after the other. It is also possible to print a summary of all elements that have been fetched from the inbox using `await inbox.summary()`. Finally, it is possible to iterate over the inbox via\n\n```python\nasync for item in inbox.iterate(max_number=3):\n    do_something(item)\n```\n\n### Proxying elements\n\nWe have already seen the difference between using `post` directly to an inbox and posting to the actor\'s outbox using `send_to_outbox`. A similar pattern applies to fetching objects. Both of these commands often have a similar result\n\n```python\nawait actor.get(object_id)\nawait actor.proxy_element(object_id)\n```\n\nHowever, they do different things:\n\n- The first `actor.get` sends a webrequest to the server `object_id` is on and retrieves it\n- The second `actor.proxy_element` sends a request to the actor\'s server for the object. This request is then either answered from the server\'s object store or by the server fetching the object. The cache behavior is up to the server. Depending of the evolution of `proxyUrl` of an Actor, more options might be added here.\n\nAs most servers don\'t support Moo-Auth-1, using `proxy_element` is the only way to obtain foreign objects, when using it.\n\n### Event Source\n\nThe event source is demonstrated in `examples/sse.py`. First, the event source will be specified in a [FEP](https://codeberg.org/fediverse/fep) to come. It provides a way to receive updates from the server, whenever a new element is added to the inbox or outbox. The basic usage is\n\n```python\nevent_source = await actor.event_source()\nasync for event in event_source:\n    if event and event.data:\n        data = json.loads(event.data)\n        do_something(data)\n```\n\nIf you plan on writing long running applications, the event source does not automatically reconnect, so you will need to implement this. [mechanical_bull](https://codeberg.org/helge/mechanical_bull) uses the event source in this way.\n',
    'author': 'Helge',
    'author_email': 'helge.krueger@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://codeberg.org/helge/bovine',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



class AcceptBuilder:
    def __init__(self, actor: str, obj: dict):
        self.actor = actor
        self.obj = obj

    def build(self) -> dict:
        return {
            "@context": "https://www.w3.org/ns/activitystreams",
            "id": self.actor + "#accepts/follows/",
            "type": "Accept",
            "actor": self.actor,
            "object": self.obj,
            "to": [self.determine_to()],
        }

    def determine_to(self) -> str:
        actor = self.obj.get("actor")

        if isinstance(actor, dict):
            actor = actor.get("id")

        return actor



import os

from .crypto import generate_public_private_key
from .http_signature import HttpSignature


def build_signature(host, method, target):
    return (
        HttpSignature()
        .with_field("(request-target)", f"{method} {target}")
        .with_field("host", host)
    )


def get_public_private_key_from_files(public_key_path, private_key_path):
    public_key = None
    private_key = None

    if os.path.exists(public_key_path):
        with open(public_key_path) as f:
            public_key = f.read()
    if os.path.exists(private_key_path):
        with open(private_key_path) as f:
            private_key = f.read()

    if public_key and private_key:
        return public_key, private_key

    public_key_pem, private_key_pem = generate_public_private_key()

    if not os.path.exists(".files"):
        os.mkdir(".files")

    with open(public_key_path, "w") as f:
        f.write(public_key_pem)

    with open(private_key_path, "w") as f:
        f.write(private_key_pem)

    return public_key_pem, private_key_pem



from .actor_builder import ActorBuilder
from .ordered_collection_builder import OrderedCollectionBuilder
from .ordered_collection_page_builder import OrderedCollectionPageBuilder


def build_actor(actor_name: str, actor_type: str = "Person"):
    return ActorBuilder(actor_name, actor_type=actor_type)


def build_ordered_collection(url: str):
    return OrderedCollectionBuilder(url)


def build_ordered_collection_page(url: str, part_of: str):
    return OrderedCollectionPageBuilder(url, part_of)



from .common import build_context


class OrderedCollectionPageBuilder:
    def __init__(self, url: str, part_of: str):
        self.context_builder = build_context()
        self.url = url
        self.items: list = []
        self.part_of = part_of
        self.next: str | None = None
        self.prev: str | None = None

    def with_items(self, items: list):
        self.items = items
        return self

    def with_next(self, url):
        self.next = url
        return self

    def with_prev(self, url):
        self.prev = url
        return self

    def build(self) -> dict:
        result = {
            "@context": self.context_builder.build(),
            "id": self.url,
            "partOf": self.part_of,
            "orderedItems": self.items,
            "type": "OrderedCollectionPage",
        }

        if self.next:
            result["next"] = self.next

        if self.prev:
            result["prev"] = self.prev

        return result



from .note_builder import NoteBuilder
from .tombstone_builder import TombstoneBuilder


def build_note(account: str, url: str, content: str) -> NoteBuilder:
    return NoteBuilder(account, content, followers=f"{account}/followers")


def tombstone(object_id):
    return TombstoneBuilder(object_id).build()



from .context_builder import ContextBuilder


def test_basic_context():
    builder = ContextBuilder()
    result = builder.build()

    assert result == "https://www.w3.org/ns/activitystreams"


def test_with_extensions():
    builder = ContextBuilder().add(
        ostatus="http://ostatus.org#", atomUri="ostatus:atomUri"
    )
    result = builder.build()

    assert result == [
        "https://www.w3.org/ns/activitystreams",
        {"ostatus": "http://ostatus.org#", "atomUri": "ostatus:atomUri"},
    ]


def test_with_additional_list_item():
    builder = ContextBuilder().add("https://w3id.org/security/v1")

    result = builder.build()

    assert result == [
        "https://www.w3.org/ns/activitystreams",
        "https://w3id.org/security/v1",
    ]


def test_absorb():
    obj = {
        "@context": ContextBuilder()
        .add(
            "https://w3id.org/security/v1",
            ostatus="http://ostatus.org#",
            atomUri="ostatus:atomUri",
        )
        .build(),
        "prop": "value",
    }

    builder = ContextBuilder()
    obj = builder.absorb(obj)

    assert "@context" not in obj

    result = builder.build()

    assert result == [
        "https://www.w3.org/ns/activitystreams",
        "https://w3id.org/security/v1",
        {"ostatus": "http://ostatus.org#", "atomUri": "ostatus:atomUri"},
    ]



from bovine.activitystreams.objects.note_builder import NoteBuilder
from bovine.activitystreams.objects.object_builder import ObjectBuilder


class ObjectFactory:
    def __init__(self, actor_information):
        self.information = actor_information

    def note(self, text):
        return NoteBuilder(
            self.information["id"], text, followers=self.information.get("followers")
        )

    def article(self):
        return ObjectBuilder(
            "Article",
            self.information["id"],
            followers=self.information.get("followers"),
        )
